
#include <stdio.h>
#include <WinSock2.h>
#include <Windows.h>
#include <detours.h>

extern "C" {
	#include <lua.h>
	#include <lauxlib.h>
	#include <lualib.h>
};

#include "steam/steamtypes.h"
#include "steam/isteamclient.h"
#include "steam/isteammatchmaking.h"
#include "steam/steam_api.h"

#include "interface.h"
#include "filesystem.h"
#include "engine/iserverplugin.h"
#include "engine/ivmodelinfo.h"
#include "studio.h"
#include "materialsystem/imaterial.h"
#include "game/server/iplayerinfo.h"
#include "eiface.h"
#include "igameevents.h"
#include "convar.h"
#include "networkstringtabledefs.h"
#include "Color.h"
#include "vstdlib/random.h"
#include "engine/IEngineTrace.h"
#include "tier2/tier2.h"

// memdbgon must be the last include file in a .cpp file!!!
#include "tier0/memdbgon.h"

// Interfaces from the engine
IGameEventManager *gameeventmanager = NULL;
IFileSystem * filesystem = NULL;

HMODULE ws2 = NULL;

lua_State * L;

typedef int (WSAAPI * proc_recvfrom)( SOCKET s, char *buf, int len, int flags, struct sockaddr *from, int *fromlen);

proc_recvfrom orig_recvfrom = NULL;

bool enable_filter = true;

int WSAAPI my_recvfrom( SOCKET s, char *buf, int len, int flags, struct sockaddr_in *from, int *fromlen )
{
	IN_ADDR sAddr;
	unsigned short sPort;

	int ret = orig_recvfrom( s, buf, len, flags, (struct sockaddr *)from, fromlen );

	// minimum valid response length is 12 bytes (header + 0.0.0.0:0), and a multiple of 6 bytes.
	if( enable_filter && ret >= 12 && ret % 6 == 0)
	{
		// The master server always fits its reply within a single datagram.
		// Check it for the response header FF FF FF FF 66 0A
		
		if( *(int *)buf == -1 && *(buf + 4) == 0x66 && *(buf + 5) == 0x0A )
		{
			//Msg( "hidefakes: Found master server response from %s:%d.\n", inet_ntoa( from->sin_addr ), from->sin_port );

			char * ptr = buf + 6;

			while( ptr < buf + ret )
			{
				sAddr = *(IN_ADDR *)ptr;				
				sPort = ntohs(*(unsigned short *)(ptr+4));
				
				lua_getglobal( L, "FilterIP" );
				lua_pushstring( L, inet_ntoa(sAddr) );
				lua_pushnumber( L, sPort );

				// two args, one result
				int res = lua_pcall( L, 2, 1, 0 );

				if( res != 0 )
				{
					const char * errmsg = lua_tostring( L, -1 );
					Warning( "hidefakes: Error in FilterIP: %s\n", errmsg );
					lua_pop( L, 1 );
				}
				else
				{
					int b = lua_toboolean( L, -1 );
					lua_pop( L, 1 );

					if( b )
					{
						// true means REMOVE this one.
						Warning( "hidefakes: BLACKLISTED SERVER %s:%d\n", inet_ntoa(sAddr), sPort );
						
						// shift the back end of the buffer forward 6 bytes.
						uint cb = (uint)((buf + ret) - (ptr + 6));

						//Msg( "buf=%d ptr=%d ret=%d cb=%d\n", buf - buf, ptr - buf, ret, cb );

						char * tmp = (char *)malloc(cb);

						memcpy( tmp, ptr + 6, cb );
						memcpy( ptr, tmp, cb );

						free( tmp );

						ret -= 6;
					}
					else
					{
						// Leave it alone.
						ptr += 6;
					}
				}				
			}
		}
	}

	return ret;
}

//---------------------------------------------------------------------------------
// Purpose: a sample 3rd party plugin class
//---------------------------------------------------------------------------------
class CHideFakesServerPlugin: public IServerPluginCallbacks, public IGameEventListener
{
public:
	CHideFakesServerPlugin();
	~CHideFakesServerPlugin();

	// IServerPluginCallbacks methods
	virtual bool			Load(	CreateInterfaceFn interfaceFactory, CreateInterfaceFn gameServerFactory );
	virtual void			Unload( void );
	virtual void			Pause( void );
	virtual void			UnPause( void );
	virtual const char     *GetPluginDescription( void );      
	virtual void			LevelInit( char const *pMapName );
	virtual void			ServerActivate( edict_t *pEdictList, int edictCount, int clientMax );
	virtual void			GameFrame( bool simulating );
	virtual void			LevelShutdown( void );
	virtual void			ClientActive( edict_t *pEntity );
	virtual void			ClientDisconnect( edict_t *pEntity );
	virtual void			ClientPutInServer( edict_t *pEntity, char const *playername );
	virtual void			SetCommandClient( int index );
	virtual void			ClientSettingsChanged( edict_t *pEdict );
	virtual PLUGIN_RESULT	ClientConnect( bool *bAllowConnect, edict_t *pEntity, const char *pszName, const char *pszAddress, char *reject, int maxrejectlen );
	virtual PLUGIN_RESULT	ClientCommand( edict_t *pEntity, const CCommand &args );
	virtual PLUGIN_RESULT	NetworkIDValidated( const char *pszUserName, const char *pszNetworkID );
	virtual void			OnQueryCvarValueFinished( QueryCvarCookie_t iCookie, edict_t *pPlayerEntity, EQueryCvarValueStatus eStatus, const char *pCvarName, const char *pCvarValue );

	// IGameEventListener Interface
	virtual void FireGameEvent( KeyValues * event );

	virtual int GetCommandIndex() { return m_iClientCommandIndex; }

private:
	int m_iClientCommandIndex;
};


// 
// The plugin is a static singleton that is exported as an interface
//
CHideFakesServerPlugin g_PhxServerPlugin;
EXPOSE_SINGLE_INTERFACE_GLOBALVAR(CHideFakesServerPlugin, IServerPluginCallbacks, INTERFACEVERSION_ISERVERPLUGINCALLBACKS, g_PhxServerPlugin );

//---------------------------------------------------------------------------------
// Purpose: constructor/destructor
//---------------------------------------------------------------------------------
CHideFakesServerPlugin::CHideFakesServerPlugin()
{
	m_iClientCommandIndex = 0;
}

CHideFakesServerPlugin::~CHideFakesServerPlugin()
{
}

//---------------------------------------------------------------------------------
// Purpose: called when the plugin is loaded, load the interface we need from the engine
//---------------------------------------------------------------------------------
bool CHideFakesServerPlugin::Load( CreateInterfaceFn interfaceFactory, CreateInterfaceFn gameServerFactory )
{
	ConnectTier1Libraries( &interfaceFactory, 1 );
	ConnectTier2Libraries( &interfaceFactory, 1 );

	//engine = (IVEngineServer*)interfaceFactory(INTERFACEVERSION_VENGINESERVER, NULL);
	gameeventmanager = (IGameEventManager *)interfaceFactory(INTERFACEVERSION_GAMEEVENTSMANAGER,NULL);
	filesystem = (IFileSystem *)interfaceFactory(FILESYSTEM_INTERFACE_VERSION, NULL);

	// get the interfaces we want to use
	if(	! ( gameeventmanager && filesystem ) )
	{
		Warning( "hidefakes: Missing some interface, cannot continue.\n" );
		Warning( "\t* gameeventmanager=0x%08x\n", gameeventmanager );
		Warning( "\t* filesystem=0x%08x\n", filesystem );

		return false;
	}

	MathLib_Init( 2.2f, 2.2f, 0.0f, 2.0f );
	ConVar_Register( 0 );

	ws2 = GetModuleHandle( "ws2_32" );

	if( !ws2 )
	{
		Msg( "hidefakes: could not get ws2\n" );
		return false;
	}

	orig_recvfrom = (proc_recvfrom)GetProcAddress( ws2, "recvfrom" );

	DetourTransactionBegin();
	DetourUpdateThread( GetCurrentThread() );
	DetourAttach( (PVOID *)(&orig_recvfrom), my_recvfrom );
	DetourTransactionCommit();

	// Create the lua state.
	L = luaL_newstate();

	luaopen_base( L );
	luaopen_math( L );
	luaopen_string( L );
	luaopen_table( L );

	// Kill some unsafe or useless globals.
	lua_pushnil( L );
	lua_setglobal( L, "print" );

	lua_pushnil( L );
	lua_setglobal( L, "dofile" );

	lua_pushnil( L );
	lua_setglobal( L, "loadstring" );

	lua_pushnil( L );
	lua_setglobal( L, "loadfile" );

	lua_pushnil( L );
	lua_setglobal( L, "require" );

	// Read and load the filtering lua script.

	FileHandle_t fh = filesystem->Open( "addons/hidefakes/serverqueryfilter.lua", "r", "MOD" );

	if( fh )
	{
		char * pstr = (char *)malloc( filesystem->Size( fh ) + 1 );

		size_t size = filesystem->Size( fh );

		memset( pstr, 0, size + 1 );

		filesystem->Read( pstr, size + 1, fh );

		int lsres = 0;

		if( ( lsres = luaL_loadstring( L, pstr )) == 0)
		{
			//Msg( " * Executing serverqueryfilter.lua...\n" );
			int ret = lua_pcall( L, 0, 0, 0 );

			if( ret != 0 )
			{
				const char * errmsg = lua_tostring( L, -1 );
				Msg( "\tError: %s\n", errmsg );
				lua_pop( L, 1 );

				return false;
			}
		}
		else
		{
			Msg( "hidefakes: error executing serverqueryfilter.lua.\n" );

			const char * errmsg = lua_tostring( L, -1 );
			Msg( "\tError: (%d) %s\n", lsres, errmsg );
			lua_pop( L, 1 );

			return false;
		}
	} 
	else
	{
		Msg( "hidefakes: could not open serverqueryfilter.lua, cannot continue.\n" );
		return false;
	}

	Msg( " * hidefakes by phenex, loaded okay.\n" );	
	
	return true;
}

//---------------------------------------------------------------------------------
// Purpose: called when the plugin is unloaded (turned off)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::Unload( void )
{
	gameeventmanager->RemoveListener( this ); // make sure we are unloaded from the event system

	ConVar_Unregister( );
	DisconnectTier2Libraries( );
	DisconnectTier1Libraries( );
}

//---------------------------------------------------------------------------------
// Purpose: called when the plugin is paused (i.e should stop running but isn't unloaded)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::Pause( void )
{
}

//---------------------------------------------------------------------------------
// Purpose: called when the plugin is unpaused (i.e should start executing again)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::UnPause( void )
{
}

//---------------------------------------------------------------------------------
// Purpose: the name of this plugin, returned in "plugin_print" command
//---------------------------------------------------------------------------------
const char *CHideFakesServerPlugin::GetPluginDescription( void )
{
	return "Hides fake servers from the server browser based on a blacklist.";
}

//---------------------------------------------------------------------------------
// Purpose: called on level start
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::LevelInit( char const *pMapName )
{
	//Msg( "Level \"%s\" has been loaded\n", pMapName );
	gameeventmanager->AddListener( this, true );
}

//---------------------------------------------------------------------------------
// Purpose: called on level start, when the server is ready to accept client connections
//		edictCount is the number of entities in the level, clientMax is the max client count
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::ServerActivate( edict_t *pEdictList, int edictCount, int clientMax )
{
}

//---------------------------------------------------------------------------------
// Purpose: called once per server frame, do recurring work here (like checking for timeouts)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::GameFrame( bool simulating )
{
}

//---------------------------------------------------------------------------------
// Purpose: called on level end (as the server is shutting down or going to a new map)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::LevelShutdown( void ) // !!!!this can get called multiple times per map change
{
	gameeventmanager->RemoveListener( this );
}

//---------------------------------------------------------------------------------
// Purpose: called when a client spawns into a server (i.e as they begin to play)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::ClientActive( edict_t *pEntity )
{
}

//---------------------------------------------------------------------------------
// Purpose: called when a client leaves a server (or is timed out)
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::ClientDisconnect( edict_t *pEntity )
{
}

//---------------------------------------------------------------------------------
// Purpose: called on 
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::ClientPutInServer( edict_t *pEntity, char const *playername )
{
}

//---------------------------------------------------------------------------------
// Purpose: called on level start
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::SetCommandClient( int index )
{
	m_iClientCommandIndex = index;
}

//---------------------------------------------------------------------------------
// Purpose: called on level start
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::ClientSettingsChanged( edict_t *pEdict )
{
}

//---------------------------------------------------------------------------------
// Purpose: called when a client joins a server
//---------------------------------------------------------------------------------
PLUGIN_RESULT CHideFakesServerPlugin::ClientConnect( bool *bAllowConnect, edict_t *pEntity, const char *pszName, const char *pszAddress, char *reject, int maxrejectlen )
{
	return PLUGIN_CONTINUE;
}

//---------------------------------------------------------------------------------
// Purpose: called when a client types in a command (only a subset of commands however, not CON_COMMAND's)
//---------------------------------------------------------------------------------
PLUGIN_RESULT CHideFakesServerPlugin::ClientCommand( edict_t *pEntity, const CCommand &args )
{
	return PLUGIN_CONTINUE;
}

//---------------------------------------------------------------------------------
// Purpose: called when a client is authenticated
//---------------------------------------------------------------------------------
PLUGIN_RESULT CHideFakesServerPlugin::NetworkIDValidated( const char *pszUserName, const char *pszNetworkID )
{
	return PLUGIN_CONTINUE;
}

//---------------------------------------------------------------------------------
// Purpose: called when a cvar value query is finished
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::OnQueryCvarValueFinished( QueryCvarCookie_t iCookie, edict_t *pPlayerEntity, EQueryCvarValueStatus eStatus, const char *pCvarName, const char *pCvarValue )
{
	//Msg( "Cvar query (cookie: %d, status: %d) - name: %s, value: %s\n", iCookie, eStatus, pCvarName, pCvarValue );
}

//---------------------------------------------------------------------------------
// Purpose: called when an event is fired
//---------------------------------------------------------------------------------
void CHideFakesServerPlugin::FireGameEvent( KeyValues * event )
{
}